/** 
 ** HIXCOLOR.C 
 **
 **  Copyright (C) 1993, David Max
 **    185 West End Ave Apt 3H
 **    New York, NY 10023
 **    max@cs.nyu.edu
 **
 **  This file is distributed under the terms listed in the document
 **  "copying.cb", available from the author at the address above.
 **  A copy of "copying.cb" should accompany this file; if not, a copy
 **  should be available from where this file was obtained.  This file
 **  may not be distributed without a verbatim copy of "copying.cb".
 **  You should also have received a copy of the GNU General Public
 **  License along with this program (it is in the file "copying");
 **  if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
 **  Cambridge, MA 02139, USA.
 **
 **  This program is distributed in the hope that it will be useful,
 **  but WITHOUT ANY WARRANTY; without even the implied warranty of
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 **  GNU General Public License for more details.
 **/

/****************************************************************************
   This is HIXCOLOR.C

   This file contains routines for driving the Diamond Speedstar 24X
   graphics card in its special 640x480 24 bit true color mode and
   its 800x600 15 bit hi color mode.

   The DOS extender understands the paging methods used by the Diamond
   Speedstar 24X and links that into the general paging mechanism it
   uses to provide virtual memory. This makes the graphics memory
   appear to be a linear range of up to 1 Mb of memory at 0xD0000000.

   The DJGPP 256 color graphics library is invoked to activate the
   automatic paging, but once the card is put into true color or 
   hi color mode, function calls to the 256 color routines will not
   work properly.

   Latest Revision: 7/9/93
   by David Max

****************************************************************************/
#include <stdlib.h>
#include <graphics.h>
#include <dos.h>
#include <memory.h>
#include "hixcolor.h"


/*---------------------------------------------------------------------------
   True Color Routines

   The 640x480 24 bit true color mode of the Diamond Speedstar 24X
   video card is accessed by mapping the memory area to an array of
   unsigned character pointers such that each line can be referenced
   individually.

---------------------------------------------------------------------------*/
pixel_24 *TrueColorBuf[_MaxY_24];

/****************************************************************************
   void GrSetMode_24(void)

   This function places the Diamond Speedstar 24X into 640x480 24 bit
   true color mode. It first calls GrSetMode to invoke the 640x480
   256 color mode in order to enable automatic paging. Then it invokes
   the 24 bit true color mode. Finally, it initializes the TrueColorBuf
   so that TrueColorBuf[i] points to the start of the ith scan line.

****************************************************************************/
void GrSetMode_24(void)
{
  union REGS reg;
  union REGS reg2;
  int i;
  
  GrSetMode(GR_width_height_graphics,640,480);
  reg.x.ax=0x0072;
  int86(0x10,&reg,&reg2);
  for (i=0;i<_MaxY_24;i++)
    TrueColorBuf[i] = (pixel_24 *)(0xd0000000+(3*i*_MaxX_24));
}



/****************************************************************************
   void GrPlot24(int x, int y, pixel_24 red, pixel_24 grn, pixel_24 blu)

   This function sets the pixel at location (x,y) to the (r,g,b) value
   specified by red, grn, and blu.

   Note: This routine checks that (x,y) refer to a valid pixel on the
         screen.

****************************************************************************/
void GrPlot24(int x, int y, pixel_24 red, pixel_24 grn, pixel_24 blu)
{
  int i;
  if ((x<0) || (x>=_MaxX_24) || (y<0) || (y>=_MaxY_24)) return;
  i = 3*x;
  TrueColorBuf[y][i] = red;
  TrueColorBuf[y][i+1] = grn;
  TrueColorBuf[y][i+2] = blu;
}



/****************************************************************************
   void GrScanLine24(int y, int x1, int x2, pixel_24 pix[])

   This function block writes a portion of a scan line.

   Inputs:  y      - scan line (0..479)
            x1     - left pixel to start at (0..639)
            x2     - right pixel to end on (x1..639)
            pix[]  - array of pixels of length 3*(x2-x1+1)

   The pixels that will be written will be (x1,y)..(x2,y). There is a
   restriction that x1 <= x2. Otherwise, the routine returns immediately.

   Note: This routine will not handle overlapping memory regions, so do
         not use this routine to do screen to screen bitblits.

         This routine checks the x and y coordinates to ensure that all
         pixels are within the screen boundaries.

****************************************************************************/
void GrScanLine24(int y, int x1, int x2, pixel_24 pix[])
{
  register int i;                  
  
  if ((x2<0) || (x1>=_MaxX_24) || (y<0) || (y>=_MaxY_24) || (x1>x2))
     return; /* scan line not on screen or pixels out of range */

  if (x2>=_MaxX_24) x2 = _MaxX_24 - 1;
  
  i = 0;              
  /* If left edge is off screen then start on the screen at x=0 with */
  /* the first pixel being pix[-x1].                                 */
  if (x1<0) {  i = -x1;
               x1 = 0;
            }                       
             
  memcpy(&(TrueColorBuf[y][3*x1]), &(pix[3*i]), (size_t)(3*(x2-x1+1)));
}

/****************************************************************************
   void GrBitBlit24(int x1, int y1, int x2, int y2, pixel_24 pix[])
   
   This function block writes a portion of the screen.

   Inputs:  x1     - upper left x pixel coordinate
            y1     - upper left y pixel coordinate   
            x2     - lower right x pixel coordinate
            y2     - lower right y pixel coordinate
            pix[]  - array of pixels values to be blitted onto the
                     screen of length (y2-y1+1)*(x2-x1+1)*3

   The pixels that will be written will be the rectangular region defined
   by (x1,y1) and (x2,y2). There is a restriction that x1<=x2 and y1<=y2.
   Otherwise, the routine returns immediately.

   Note: This routine will not handle overlapping memory regions, so do
         not use this routine to do screen to screen bitblits.

         This routine does not explicitly handle off-screen coordinates,
         however, the scan line routine will properly clip the scan lines
         to the screen boundaries.

****************************************************************************/
void GrBitBlit24(int x1, int y1, int x2, int y2, pixel_24 pix[])
{
  register int i,ScanLineLength;

  if (x2<x1) return;                 /* invalid rectangle x2<x1 */

  ScanLineLength = 3*(x2-x1+1);
  for(i=y1; i<=y2; i++)              /* implicit check for y2<y1 */
     GrScanLine24(i,x1,x2,&(pix[(i-y1)*ScanLineLength]));
}



/****************************************************************************
   void GrLine24(int x1, int y1, int x2, int y2, 
                 pixel_24 red, pixel_24 grn, pixel_24 blu)

   This function draws a line from (x1,y1) to (x2,y2) in the color
   specified by red, grn, and blu.

   Note: This routine does NOT perform line clipping.

****************************************************************************/
void GrLine24(int x1, int y1, int x2, int y2, 
              pixel_24 red, pixel_24 grn, pixel_24 blu)
{
  int dx, dy, sx, sy;
  int count;
  int brc, brmax;

  sx = sy = 1;
  dx = x2 - x1;
  dy = y2 - y1;
  if (dx < 0)
  {
    dx *= -1;
    sx *= -1;
  }
  if (dy < 0)
  {
    dy *= -1;
    sy *= -1;
  }
  if (dx > dy)
  {
    brmax = dx;
    brc = dx / 2;
    GrPlot24(x1, y1, red, grn, blu);
    for (count = dx; count; count--)
    {
      x1 += sx;
      brc += dy;
      if (brc > brmax)
      {
        brc -= dx;
        y1 += sy;
      }
      GrPlot24(x1, y1, red, grn, blu);
    }
  }
  else
  {
    brmax = dy;
    brc = dy / 2;
    GrPlot24(x1, y1, red, grn, blu);
    for (count = dy; count; count--)
    {
      y1 += sy;
      brc += dx;
      if (brc > brmax)
      {
        brc -= dy;
        x1 += sx;
      }
      GrPlot24(x1, y1, red, grn, blu);
    }
  }
}


/*---------------------------------------------------------------------------
   Hi Color Routines

   The 800x600 15 bit hi color mode of the Diamond Speedstar 24X
   video card is accessed by mapping the memory area to an array of
   unsigned short integer pointers such that each line can be referenced
   individually.

---------------------------------------------------------------------------*/
pixel_HI *HiColorBuf[_MaxY_HI];

/****************************************************************************
   void GrSetMode_HI(void)

   This function places the Diamond Speedstar 24X into 800x600 15 bit
   hi color mode. It first calls GrSetMode to invoke the 640x480
   256 color mode in order to enable automatic paging. Then it invokes
   the 15 bit hi color mode. Finally, it initializes the HiColorBuf
   so that HiColorBuf[i] points to the start of the ith scan line.

****************************************************************************/
void GrSetMode_HI(void)
{
  union REGS reg;
  union REGS reg2;
  int i;
  
  GrSetMode(GR_width_height_graphics,640,480);
  reg.x.ax=0x0063;
  int86(0x10,&reg,&reg2);
  for (i=0;i<_MaxY_HI;i++)
    HiColorBuf[i] = (pixel_HI *)(0xd0000000+(2*i*_MaxX_HI));
}


/****************************************************************************
   void GrPlotHI(int x, int y, pixel_24 red, pixel_24 grn, pixel_24 blu)

   This function sets the pixel at location (x,y) to the (r,g,b) value
   specified by red, grn, and blu.

   Note: This routine checks that (x,y) refer to a valid pixel on the
         screen. Also note that red, grn, and blu should be in the 
         range of 0..31.

****************************************************************************/
void GrPlotHI(int x, int y, pixel_24 red, pixel_24 grn, pixel_24 blu)
{     
  register pixel_HI c;
  if ((x<0) || (x>=_MaxX_HI) || (y<0) || (y>=_MaxY_HI))
    return;
  c = (((red<<5)|grn)<<5)|blu;
  HiColorBuf[y][x] = c;
}

/****************************************************************************
   void GrPlotHI2(int x, int y, pixel_HI c)

   This function sets the pixel at location (x,y) to the color
   specified by c. c is a two byte short int with a bit pattern of
   0RRRRRGGGGGBBBBB specifying the red, green, and blue components of
   the desired color.

   Note: This routine checks that (x,y) refer to a valid pixel on the
         screen. 

****************************************************************************/
void GrPlotHI2(int x, int y, pixel_HI c)
{     
  if ((x<0) || (x>=_MaxX_HI) || (y<0) || (y>=_MaxY_HI))
    return;
  HiColorBuf[y][x] = c;
}

/****************************************************************************
   void GrScanLineHI(int y, int x1, int x2, pixel_HI pix[])

   This function block writes a portion of a scan line.

   Inputs:  y      - scan line (0..599)
            x1     - left pixel to start at (0..799)
            x2     - right pixel to end on (x1..799)
            pix[]  - array of pixels of length x2-x1+1

   The pixels that will be written will be (x1,y)..(x2,y). There is a
   restriction that x1 <= x2. Otherwise, the routine returns immediately.

   Note: This routine will not handle overlapping memory regions, so do
         not use this routine to do screen to screen bitblits.

         This routine checks the x and y coordinates to ensure that all
         pixels are within the screen boundaries.

****************************************************************************/
void GrScanLineHI(int y, int x1, int x2, pixel_HI pix[])
{
  register int i;
  if ((x2<0) || (x1>=_MaxX_HI) || (y<0) || (y>=_MaxY_HI) || (x1>x2))
     return; /* scan line not on screen or pixels out of range */ 
     
  if (x2>=_MaxX_HI) x2 = _MaxX_HI - 1;

  i=0;
  /* If left edge is off screen then start on the screen at x=0 with */
  /* the first pixel being pix[-x1].                                 */
  if (x1<0) {  i = -x1;
               x1 = 0;
            }

  memcpy(&(HiColorBuf[y][x1]), &(pix[i]), (size_t)(2*(x2-x1+1)));
}

/****************************************************************************
   void GrBitBlitHI(int x1, int y1, int x2, int y2, pixel_HI pix[])
   
   This function block writes a portion of the screen.

   Inputs:  x1     - upper left x pixel coordinate
            y1     - upper left y pixel coordinate   
            x2     - lower right x pixel coordinate
            y2     - lower right y pixel coordinate
            pix[]  - array of pixels values to be blitted onto the
                     screen of length (y2-y1+1)*(x2-x1+1)

   The pixels that will be written will be the rectangular region defined
   by (x1,y1) and (x2,y2). There is a restriction that x1<=x2 and y1<=y2.
   Otherwise, the routine returns immediately.

   Note: This routine will not handle overlapping memory regions, so do
         not use this routine to do screen to screen bitblits.

         This routine does not explicitly handle off-screen coordinates,
         however, the scan line routine will properly clip the scan lines
         to the screen boundaries.

****************************************************************************/
void GrBitBlitHI(int x1, int y1, int x2, int y2, pixel_HI pix[])
{
  int i,ScanLineLength;

  if (x2<x1) return;                 /* invalid rectangle x2<x1 */

  ScanLineLength = x2-x1+1;
  for(i=y1; i<=y2; i++)              /* implicit check for y2<y1 */
     GrScanLineHI(i,x1,x2,&(pix[(i-y1)*ScanLineLength]));
}


/****************************************************************************
   void GrLineHI(int x1, int y1, int x2, int y2, 
                 pixel_24 red, pixel_24 grn, pixel_24 blu)

   This function draws a line from (x1,y1) to (x2,y2) in the color
   specified by red, grn, and blu.

   Note: This routine does NOT perform line clipping.

****************************************************************************/
void GrLineHI(int x1, int y1, int x2, int y2, 
              pixel_24 red, pixel_24 grn, pixel_24 blu)
{
  pixel_HI c;

  c = (((red<<5)|grn)<<5)|blu;
  GrLineHI2(x1, y1, x2, y2, c);
}

/****************************************************************************
   void GrLineHI2(int x1, int y1, int x2, int y2, pixel_HI c)

   This function draws a line from (x1,y1) to (x2,y2) in the color
   specified by c.

   Note: This routine does NOT perform line clipping.

****************************************************************************/
void GrLineHI2(int x1, int y1, int x2, int y2, pixel_HI c)
{
  int dx, dy, sx, sy;
  int count;
  int brc, brmax;
  
  sx = sy = 1;
  dx = x2 - x1;
  dy = y2 - y1;
  if (dx < 0)
  {
    dx *= -1;
    sx *= -1;
  }
  if (dy < 0)
  {
    dy *= -1;
    sy *= -1;
  }
  if (dx > dy)
  {
    brmax = dx;
    brc = dx / 2;
    GrPlotHI2(x1, y1, c);
    for (count = dx; count; count--)
    {
      x1 += sx;
      brc += dy;
      if (brc > brmax)
      {
        brc -= dx;
        y1 += sy;
      }
      GrPlotHI2(x1, y1, c);
    }
  }
  else
  {
    brmax = dy;
    brc = dy / 2;
    GrPlotHI2(x1, y1, c);
    for (count = dy; count; count--)
    {
      y1 += sy;
      brc += dx;
      if (brc > brmax)
      {
        brc -= dy;
        x1 += sx;
      }
      GrPlotHI2(x1, y1, c);
    }
  }
}

