/* Copyright (C) 1989, 1990 Aladdin Enterprises.  All rights reserved.
   Distributed by Free Software Foundation, Inc.

This file is part of Ghostscript.

Ghostscript is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the Ghostscript General Public License for full details.

Everyone is granted permission to copy, modify and redistribute
Ghostscript, but only under the conditions described in the Ghostscript
General Public License.  A copy of this license is supposed to have been
given to you along with Ghostscript so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.  */

/* gdevegad(ir).c */
/* EGA driver for Ghostscript using direct access to frame buffer */
#define GS_EGA_DEVICE gs_ega_device
#define DEVICE_NAME "ega"
#include "gdevega.h"
#define USE_ASM 1			/* use assembly language */

/* Type and macro for frame buffer pointers. */
/* Note: requires (raster_x & 15) == 0. */
/*** Intimately tied to the 80x86 (x<2) addressing architecture. ***/
typedef byte far *fb_ptr;
#define mk_fb_ptr(x, y)\
  ((fb_ptr)MK_FP(regen + (y) * (raster_x >> 4), (x) >> 3))
/* Define the largest height that can be processed */
/* within a single 64K segment.  If screen_size_y > max_rop_height, */
/* we may have to break up operations into pieces. */
#define max_rop_height (0xffff / raster_x)

/* Structure for operation parameters. */
/* Note that this structure is known to assembly code. */
/* Not all parameters are used for every operation. */
typedef struct rop_params_s {
	fb_ptr dest;			/* pointer to frame buffer */
	int draster;			/* raster of frame buffer */
	byte far *src;			/* pointer to source data */
	int sraster;			/* source raster */
	int width;			/* width in bytes */
	int height;			/* height in scan lines */
	int shift;			/* amount to right shift source */
	int invert;			/* 0 or -1 to invert source */
	int data;			/* data for fill */
} rop_params;
typedef rop_params _ss *rop_ptr;

/* Assembly language routines */

#if USE_ASM
void memsetcol(P1(rop_ptr)); /* dest, draster, height, data */
#else
#define memsetcol cmemsetcol
private void
cmemsetcol(rop_ptr rop)
{	byte far *addr = rop->dest;
	int yc = rop->height;
	while ( yc-- )
	 { byte discard = *addr;
	   *addr = rop->data;
	   addr += rop->draster;
	 }
}
#endif

#if USE_ASM
void memsetrect(P1(rop_ptr)); /* dest, draster, width, height, data */
#else
#define memsetrect cmemsetrect
private void
cmemsetrect(rop_ptr rop)
{	byte far *addr = rop->dest;
	int yc = rop->height;
	while ( yc-- )
	 { int cnt = rop->width;
	   while ( cnt-- ) *addr++ = rop->data;
	   addr += rop->draster - rop->width;
	 }
}
#endif

#if USE_ASM
void memrwcol(P1(rop_ptr)); /* dest, draster, src, sraster, height, shift, invert */
#else
#define memrwcol cmemrwcol
private void
cmemrwcol(rop_ptr rop)
{	byte far *dp = rop->dest, *sp = rop->src;
	int yc = rop->height;
	int shift = rop->shift;
	while ( yc-- )
	 { byte discard = *dp;
	   *dp = ((*sp >> shift) + (*sp << (8 - shift))) ^ rop->invert;
	   dp += rop->draster, sp += rop->sraster;
	 }
}
#endif

#if USE_ASM
void memrwcol2(P1(rop_ptr)); /* dest, draster, src, sraster, height, shift, invert */
#else
#define memrwcol2 cmemrwcol2
private void
cmemrwcol2(rop_ptr rop)
{	byte far *dp = rop->dest, *sp = rop->src;
	int yc = rop->height;
	int shift = rop->shift;
	while ( yc-- )
	 { byte discard = *dp;
	   *dp = ((sp[1] >> shift) + (*sp << (8 - shift))) ^ rop->invert;
	   dp += rop->draster, sp += rop->sraster;
	 }
}
#endif

/* Forward definitions */
void fill_rectangle(P4(rop_ptr, int, int, int));

/* Define the device port and register numbers, and the regen map base */
#define out2(port, index, data)\
  (outportb(port, index), outportb((port)+1, data))
#define seq_addr 0x3c4
#define s_map 2
#define set_s_map(mask) out2(seq_addr, s_map, mask)
#define graph_addr 0x3ce
#define g_const 0			/* set/reset */
#define set_g_const(color) out2(graph_addr, g_const, color)
#define g_const_map 1			/* enable set/reset */
#define set_g_const_map(map) out2(graph_addr, g_const_map, map)
#define g_function 3
#define set_g_function(func) out2(graph_addr, g_function, func)
#  define gf_WRITE 0
#  define gf_AND 8
#  define gf_OR 0x10
#  define gf_XOR 0x18
#define g_mode 5
#define set_g_mode(mode) out2(graph_addr, g_mode, mode)
#  define gm_DATA 0
#  define gm_FILL 2
#define g_mask 8
#define set_g_mask(mask) out2(graph_addr, g_mask, mask)
#define regen 0xa000

/* Clean up after writing */
#define dot_end()\
  set_g_mask(0xff)			/* all bits on */

/* Initialize the EGA for graphics mode */
int
ega_open(gx_device *dev)
{	if ( ega_save_mode < 0 ) ega_save_mode = ega_get_mode();
	ega_set_mode(graphics_video_mode);
	set_s_map(-1);			/* enable all maps */
	return 0;
}

/* Write a dot using the EGA color codes. */
/* This doesn't have to be efficient. */
int
ega_write_dot(gx_device *dev, int x, int y, gx_color_index color)
{	byte data = (byte)color;
	return ega_copy_color(dev, &data, 1, 1, x, y, 1, 1);
}

/* Macro for testing bit-inclusion */
#define bit_included_in(x,y) !((x)&~(y))

/* Copy a monochrome bitmap.  The colors are given explicitly. */
/* Color = gx_no_color_index means transparent (no effect on the image). */
int
ega_copy_mono(gx_device *dev, byte *base, int sourcex, int raster,
  int x, int y, int w, int h, gx_color_index izero, gx_color_index ione)
{	rop_params params;
#define czero (int)izero
#define cone (int)ione
	int dleft, count;
	byte mask, rmask;
	fb_ptr save_dest;
	int other_color = -1;
	validate_rect();
#if screen_size_y > max_rop_height
	while ( h > max_rop_height )
	   {	int code = ega_copy_mono(dev, base, sourcex, raster,
				x, y, w, max_rop_height, izero, ione);
		if ( code != 0 ) return code;
		base = (byte *)((byte huge *)base +
				raster * (long)max_rop_height);
		y += max_rop_height;
		h -= max_rop_height;
	   }
#endif
	params.dest = mk_fb_ptr(x, y);
	params.draster = raster_x;
	params.src = base + (sourcex >> 3);
	params.sraster = raster;
	params.height = h;
	params.shift = (x - sourcex) & 7;
	/* Analyze the 16 possible cases: each of izero and ione may be */
	/* 0, 0xf, transparent, or some other color. */
	switch ( czero )
	   {
	case no_color:
		switch ( cone )
		   {
		default:		/* (T, other) */
			/* Must do 2 passes */
			other_color = cone;
			save_dest = params.dest;
			/* falls through */
		case 0:			/* (T, 0) */
			set_g_function(gf_AND);
			params.invert = -1;
			break;
		case 0xf:		/* (T, 0xf) */
			set_g_function(gf_OR);
			params.invert = 0;
			break;
		case no_color:		/* (T, T) */
			return 0;	/* nothing to do */
		   }
		break;
	case 0:
		params.invert = 0;
		switch ( cone )
		   {
		default:		/* (0, other) */
			set_g_const(0);
			set_g_const_map(cone ^ 0xf);
			/* falls through */
		case 0xf:		/* (0, 0xf) */
			break;
		case no_color:		/* (0, T) */
			set_g_function(gf_AND);
			break;
		   }
		break;
	case 0xf:
		params.invert = -1;
		switch ( cone )
		   {
		case 0:			/* (0xf, 0) */
			break;
		default:		/* (0xf, other) */
			set_g_const(0xf);
			set_g_const_map(cone);
			break;
		case no_color:		/* (0xf, T) */
			set_g_function(gf_OR);
			/* falls through */
		   }
		break;
	default:
		switch ( cone )
		   {
		default:		/* (other, not T) */
			if ( bit_included_in(czero, cone) )
			   {	set_g_const(czero);
				set_g_const_map(czero ^ cone ^ 0xf);
				params.invert = 0;
				break;
			   }
			else if ( bit_included_in(cone, czero) )
			   {	set_g_const(cone);
				set_g_const_map(cone ^ czero ^ 0xf);
				params.invert = -1;
				break;
			   }
			/* No way around it, fill with one color first. */
			save_dest = params.dest;
			fill_rectangle((rop_ptr)&params, x & 7, w, cone);
			params.dest = save_dest;
			set_g_function(gf_XOR);
			set_s_map(czero ^ cone);
			other_color = -2;	/* must reset s_map at end */
			params.invert = -1;
			break;
		case no_color:		/* (other, T) */
			/* Must do 2 passes */
			other_color = czero;
			save_dest = params.dest;
			set_g_function(gf_AND);
			params.invert = 0;
			break;
		   }
		break;
	   }
	/* Actually copy the bits. */
	dleft = 8 - (x & 7);
	mask = 0xff >> (8 - dleft);
	count = w - dleft;
	if ( count < 0 )
		mask -= mask >> w,
		rmask = 0;
	else
		rmask = 0xff00 >> (count & 7);
	/* params: dest, src, sraster, height, shift, invert */
	/* Smashes params.src, params.dest, count. */
copy:	set_g_mask(mask);
	if ( params.shift == 0 )	/* optimize the aligned case */
	   {	/* Do left column */
		memrwcol((rop_ptr)&params);
		/* Do center */
		if ( (count -= 8) >= 0 )
		   {	set_g_mask(0xff);
			do
			   {	params.src++, params.dest++;
				memrwcol((rop_ptr)&params);
			   }
			while ( (count -= 8) >= 0 );
		   }
		/* Do right column */
		if ( rmask )
		   {	params.src++, params.dest++;
			set_g_mask(rmask);
			memrwcol((rop_ptr)&params);
		   }
	   }
	else
	   {	/* Do left column */
		int sleft = 8 - (sourcex & 7);
		if ( sleft >= dleft )
		   {	/* Source fits in one byte */
			memrwcol((rop_ptr)&params);
		   }
		else if ( w <= sleft )
		   {	/* Source fits in one byte, thin case */
			memrwcol((rop_ptr)&params);
			goto fin;
		   }
		else
		   {	memrwcol2((rop_ptr)&params);
			params.src++;
		   }
		/* Do center */
		if ( (count -= 8) >= 0 )
		   {	set_g_mask(0xff);
			do
			   {	params.dest++;
				memrwcol2((rop_ptr)&params);
				params.src++;
			   }
			while ( (count -= 8) >= 0 );
		   }
		/* Do right column */
		if ( rmask )
		   {	set_g_mask(rmask);
			params.dest++;
			if ( count + 8 <= params.shift )
				memrwcol((rop_ptr)&params);
			else
				memrwcol2((rop_ptr)&params);
		   }
	   }
fin:	if ( other_color != -1 )
	   {	if ( other_color >= 0 )
		   {	/* Do the second pass on (T, other) or (other, T). */
			count = w - dleft;
			params.src = base + (sourcex >> 3);
			params.dest = save_dest;
			params.invert ^= -1;
			set_s_map(other_color);
			set_g_function(gf_OR);
			other_color = -2;
			goto copy;
		   }
		else
		   {	/* Finished second pass, restore s_map */
			set_s_map(-1);
		   }
	   }
	set_g_function(gf_WRITE);
	set_g_const_map(0);
	dot_end();
	return 0;
#undef czero
#undef cone
}

/* Copy a color pixelmap.  This is just like a bitmap, */
/* except that each pixel takes 4 bits instead of 1. */
int
ega_copy_color(gx_device *dev, byte *base, int sourcex, int raster,
  int x, int y, int w, int h)
{	byte *line = base + (sourcex >> 1);
	unsigned lmask = 0x80 >> (x & 7);
	int bitx = sourcex & 1;
	int ex = w + bitx;
	fb_ptr fbline;
	validate_rect();
#if screen_size_y > max_rop_height
	while ( h > max_rop_height )
	   {	int code = ega_copy_color(dev, base, sourcex, raster,
				x, y, w, max_rop_height);
		if ( code != 0 ) return code;
		base = (byte *)((byte huge *)base +
				raster * (long)max_rop_height);
		y += max_rop_height;
		h -= max_rop_height;
	   }
#endif
	fbline = mk_fb_ptr(x, y);
	set_g_mode(gm_FILL);
	outportb(graph_addr, g_mask);	/* first half of set_g_mask */
	while ( --h >= 0 )
	{	byte *bptr = line;
		int px = bitx;
		/* Turbo C will put an unsigned in a register, */
		/* but not a byte! */
		register unsigned mask = lmask;
		fb_ptr fbptr = fbline;
		do
		   {	byte discard = *fbptr;	/* latch frame buffer data */
			outportb(graph_addr+1, mask);	/* 2nd half of set_g_mask */
			*fbptr = (px & 1 ? *bptr++ : *bptr >> 4);
			if ( (mask >>= 1) == 0 )
				mask = 0x80, fbptr++;
		   }
		while ( ++px < ex );
		line += raster;
		fbline += raster_x;
	}
	set_g_mode(gm_DATA);
	dot_end();
	return 0;
}

/* Fill a rectangle. */
int
ega_fill_rectangle(gx_device *dev, int x, int y, int w, int h,
  gx_color_index color)
{	rop_params params;
	validate_rect();
#if screen_size_y > max_rop_height
	while ( h > max_rop_height )
	   {	int code = ega_fill_rectangle(dev,
			x, y, w, max_rop_height, color);
		if ( code != 0 ) return code;
		y += max_rop_height;
		h -= max_rop_height;
	   }
#endif
	params.dest = mk_fb_ptr(x, y);
	params.draster = raster_x;
	params.height = h;
	fill_rectangle((rop_ptr)&params, x & 7, w, (int)color);
	dot_end();
	return 0;
}

/* Tile a rectangle.  Note that the two colors must both be supplied, */
/* i.e. neither one can be gx_no_color_index (transparent): */
/* a transparent color means that the tile is colored, not a mask. */
int
ega_tile_rectangle(gx_device *dev, gx_bitmap *tile,
  int x, int y, int w, int h, gx_color_index czero, gx_color_index cone)
#define zero (int)czero
#define one (int)cone
{	rop_params params;
	int xmod, width_bytes;
	int tile_height = tile->height;
	int xbit;
	int lcount;
	int mask, rmask;
	byte narrow;
	byte again;
	int const_bits, maps;
	int ymod, yleft;
#if screen_size_y > max_rop_height
	while ( h > max_rop_height )
	   {	int mod_height = max_rop_height / tile_height * tile_height;
		int code = ega_tile_rectangle(dev, tile,
			x, y, w, mod_height, czero, cone);
		if ( code != 0 ) return code;
		y += mod_height;
		h -= mod_height;
	   }
#endif
	if ( tile->width & 7 ) return -1;	/* can't do it */
	if ( one == -1 || zero == -1 ) return -1;	/* can't do colors */
	/* Following is similar to aligned case of copy_mono */	
	validate_rect();
	params.dest = mk_fb_ptr(x, y);
	params.draster = raster_x;
	params.sraster = tile->raster;
	params.shift = 0;
	xbit = x & 7;
	/* Set up the graphics registers */
	const_bits = (zero ^ one) ^ 0xf;
	if ( const_bits )
	   {	set_g_const(zero);	/* either color will do */
		set_g_const_map(const_bits);
	   }
	if ( (maps = zero & ~one) != 0 )
	   {	set_s_map(maps += const_bits);
		params.invert = -1;
		again = one & ~zero;
	   }
	else
	   {	maps = one & ~zero;
		set_s_map(maps += const_bits);
		params.invert = 0;
		again = 0;
	   }
	xmod = (x % tile->width) >> 3;
	width_bytes = tile->width >> 3;
	mask = 0xff >> xbit;
	if ( w + xbit <= 8 )
		mask -= mask >> w,
		rmask = 0,
		narrow = 1;
	else
		rmask = (0xff00 >> ((w + x) & 7)) & 0xff,
		w += xbit - 8,
		narrow = 0;
	ymod = y % tile_height;
tile:	yleft = tile_height - ymod;
	params.src = tile->data + ymod * params.sraster + xmod;
	lcount = h;
	if ( narrow )			/* Optimize narrow case */
	   {	set_g_mask(mask);
		if ( lcount > yleft )
		   {	params.height = yleft;
			memrwcol((rop_ptr)&params);
			params.dest += yleft * raster_x;
			params.src = tile->data + xmod;
			params.height = tile_height;
			lcount -= yleft;
			while ( lcount >= tile_height )
			   {	memrwcol((rop_ptr)&params);
				params.dest += tile_height * raster_x;
				lcount -= tile_height;
			   }
		   }
		if ( lcount )
		   {	params.height = lcount;
			memrwcol((rop_ptr)&params);
		   }
	   }
	else
	   {	fb_ptr line = params.dest;
		while ( 1 )
		   {	int xoff = xmod;
			int count = w;
			params.height = (lcount > yleft ? yleft : lcount);
			/* Do first byte */
			set_g_mask(mask);
			memrwcol((rop_ptr)&params);
			/* Do full bytes */
			if ( (count -= 8) >= 0 )
			   {	set_g_mask(0xff);
				do
				   {	if ( ++xoff == width_bytes )
						xoff = 0,
						params.src -= params.sraster;
					++params.src, ++params.dest;
					memrwcol((rop_ptr)&params);
				   }
				while ( (count -= 8) >= 0 );
			   }
			/* Do last byte */
			if ( rmask )
			   {	if ( ++xoff == width_bytes )
					xoff = 0,
					params.src -= params.sraster;
				set_g_mask(rmask);
				++params.src, ++params.dest;
				memrwcol((rop_ptr)&params);
			   }
			if ( (lcount -= params.height) == 0 ) break;
			params.dest = line += params.height * raster_x;
			params.src = tile->data + xmod;
			yleft = tile_height;
		   }
	   }
	/* Now do the second color if needed */
	if ( again )
	   {	maps = again + const_bits;
		set_s_map(maps);
		again = 0;
		params.dest = mk_fb_ptr(x, y);
		params.invert = 0;
		goto tile;
	   }
	if ( maps != 0xf )
		set_s_map(-1);
	if ( const_bits )
		set_g_const_map(0);
	dot_end();
	return 0;
}

/* ------ Internal routines ------ */

/* Fill a rectangle specified by pointer into frame buffer, */
/* starting bit within byte, width, and height. */
/* Smashes rop->dest. */
private void
fill_rectangle(register rop_ptr rop, int bit, int w, int color)
  /* rop: dest, draster, height */
{	static byte rmask_tab[9] =
	   {	0x00, 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff
	   };
	set_g_const(color);
	set_g_const_map(0xf);
	if ( bit + w <= 8 )
	   {	/* Less than one byte */
		set_g_mask(rmask_tab[w] >> bit);
		memsetcol(rop);
	   }
	else
	   {	int byte_count;
		byte right_mask;
		if ( bit )
		   {	set_g_mask(0xff >> bit);
			memsetcol(rop);
			rop->dest++;
			w += bit - 8;
		   }
		if ( (byte_count = w >> 3) != 0 )
		   {	set_g_mask(0xff);	/* all bits */
			rop->width = byte_count;
			memsetrect(rop);
			rop->dest += byte_count;
		   }
		if ( (right_mask = rmask_tab[w & 7]) != 0 )
		   {	set_g_mask(right_mask);
			memsetcol(rop);
		   }
	   }
	set_g_const_map(0);
}
