/* chmod -- change permission modes of files
   Copyright (C) 1989, 1990 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Usage: chmod [-Rcdfv] mode file...
          mode is [ugoa...][[+-=][rwxXstugo...]...][,...] or octal number.

   Options:
   -R	Recursively change modes of directory contents.
   -c	Verbosely describe only files whose modes actually change.
   -d	Dereference symbolic links (recursively change the modes of
	directories pointed to by symbolic links).
   -f	Do not print error messages about files.
   -v	Verbosely describe changed modes.

   David MacKenzie <djm@ai.mit.edu> */

#include <stdio.h>
#include <sys/types.h>
#include "getopt.h"
#include "modechange.h"
#include "system.h"

#ifdef STDC_HEADERS
#include <errno.h>
#include <stdlib.h>
#else
char *malloc ();
char *realloc ();

extern int errno;
#endif

int lstat ();
int stat ();

char *savedir ();
char *xmalloc ();
char *xrealloc ();
int change_file_mode ();
int change_dir_mode ();
void describe_change ();
void error ();
void mode_string ();
void usage ();

typedef enum
{
  false = 0, true = 1
} boolean;

/* The name the program was run with. */
char *program_name;

/* If true, change the modes of directories recursively. */
boolean recurse;

/* If true, force silence (no error messages). */
boolean force_silent;

/* If true, describe the modes we set. */
boolean verbose;

/* If true, describe only modes that change. */
boolean changes_only;

/* A pointer to either lstat or stat. */
int (*xstat) ();

/* Parse the ASCII mode given on the command line into a linked list
   of `struce mode_change' and apply that to each file argument. */

void
main (argc, argv)
     int argc;
     char **argv;
{
  extern int optind;
  struct mode_change *changes;
  int errors = 0;
  int modeind = 0;		/* Index of the mode argument in `argv'. */
  int c;

  program_name = argv[0];
  recurse = force_silent = verbose = changes_only = false;
  xstat = lstat;

  while ((c = getopt (argc, argv, "RcdfvrwxXstugoa,+-=")) != EOF)
    {
      switch (c)
	{
	case 'r':
	case 'w':
	case 'x':
	case 'X':
	case 's':
	case 't':
	case 'u':
	case 'g':
	case 'o':
	case 'a':
	case ',':
	case '+':
	case '-':
	case '=':
	  if (modeind != 0 && modeind != optind - 1)
	    error (1, 0, "invalid mode");
	  modeind = optind - 1;
	  break;
	case 'R':
	  recurse = true;
	  break;
	case 'c':
	  verbose = true;
	  changes_only = true;
	  break;
	case 'd':
	  xstat = stat;
	  break;
	case 'f':
	  force_silent = true;
	  break;
	case 'v':
	  verbose = true;
	  break;
	default:
	  usage ();
	}
    }

  if (modeind == 0)
    modeind = optind++;
  if (optind >= argc)
    usage ();

  changes = mode_compile (argv[modeind], MODE_MASK_EQUALS | MODE_MASK_PLUS);
  if (changes == MODE_INVALID)
    error (1, 0, "invalid mode");
  else if (changes == MODE_MEMORY_EXHAUSTED)
    error (1, 0, "virtual memory exhausted");

  for (; optind < argc; ++optind)
    errors |= change_file_mode (argv[optind], changes);

  exit (errors);
}

/* Change the mode of `file' according to the list of operations `changes'.
   Return 0 if successful, 1 if errors occurred. */

int
change_file_mode (file, changes)
     char *file;
     struct mode_change *changes;
{
  struct stat file_stats;
  unsigned short newmode;
  int errors = 0;

  if ((*xstat) (file, &file_stats))
    {
      if (force_silent == false)
	error (0, errno, "%s", file);
      return 1;
    }
#ifdef S_IFLNK
  if ((file_stats.st_mode & S_IFMT) == S_IFLNK)
    return 0;
#endif

  newmode = mode_adjust (file_stats.st_mode, changes);

  if (newmode != (file_stats.st_mode & 07777))
    {
      if (verbose)
	describe_change (file, newmode, 1);
      if (chmod (file, (int) newmode))
	{
	  if (force_silent == false)
	    error (0, errno, "%s", file);
	  errors = 1;
	}
    }
  else if (verbose && changes_only == false)
    describe_change (file, newmode, 0);

  if (recurse && (file_stats.st_mode & S_IFMT) == S_IFDIR)
    errors |= change_dir_mode (file, changes, &file_stats);
  return errors;
}

/* Recursively change the modes of the files in directory `dir'
   according to the list of operations `changes'.
   `statp' points to the results of lstat or stat on `dir'.
   Return 0 if successful, 1 if errors occurred. */

int
change_dir_mode (dir, changes, statp)
     char *dir;
     struct mode_change *changes;
     struct stat *statp;
{
  char *name_space, *namep;
  char *path;			/* Full path of each entry to process. */
  unsigned dirlength;		/* Length of `dir' and '\0'. */
  unsigned filelength;		/* Length of each pathname to process. */
  unsigned pathlength;		/* Bytes allocated for `path'. */
  int errors = 0;

  errno = 0;
  name_space = savedir (dir, statp->st_size);
  if (name_space == NULL)
    {
      if (errno)
	{
	  if (force_silent == false)
	    error (0, errno, "%s", dir);
	  return 1;
	}
      else
	error (1, 0, "virtual memory exhausted");
    }

  dirlength = strlen (dir) + 1;	/* + 1 is for the trailing '/'. */
  pathlength = dirlength + 1;
  /* Give `path' a dummy value; it will be reallocated before first use. */
  path = xmalloc (pathlength);
  strcpy (path, dir);
  path[dirlength - 1] = '/';

  for (namep = name_space; *namep; namep += filelength - dirlength)
    {
      filelength = dirlength + strlen (namep) + 1;
      if (filelength > pathlength)
	{
	  pathlength = filelength * 2;
	  path = xrealloc (path, pathlength);
	}
      strcpy (path + dirlength, namep);
      errors |= change_file_mode (path, changes);
    }
  free (path);
  free (name_space);
  return errors;
}

/* Tell the user the mode `mode' that file `file' has been set to;
   if `changed' is zero, `file' had that mode already. */

void
describe_change (file, mode, changed)
     char *file;
     unsigned short mode;
     int changed;
{
  char perms[11];		/* "-rwxrwxrwx" ls-style modes. */

  mode_string (mode, perms);
  perms[10] = '\0';		/* `mode_string' does not null terminate. */
  if (changed)
    printf ("mode of %s changed to %04o (%s)\n",
	    file, mode & 07777, &perms[1]);
  else
    printf ("mode of %s retained as %04o (%s)\n",
	    file, mode & 07777, &perms[1]);
}

/* Allocate `n' bytes of memory dynamically, with error checking.  */

char *
xmalloc (n)
     unsigned n;
{
  char *p;

  p = malloc (n);
  if (p == 0)
    error (1, 0, "virtual memory exhausted");
  return p;
}

char *
xrealloc (p, n)
     char *p;
     unsigned n;
{
  p = realloc (p, n);
  if (p == 0)
    error (1, 0, "virtual memory exhausted");
  return p;
}

void
usage ()
{
  fprintf (stderr, "\
Usage: %s [-Rcdfv] mode file...\n\
       mode is [ugoa...][[+-=][rwxXstugo...]...][,...] or octal number\n",
	   program_name);
  exit (1);
}
